import pytest  # type: ignore

from abacus.api import Chart, ChartCommand
from abacus.engine.accounts import ContraIncome


@pytest.fixture
def chart():
    return ChartCommand(Chart())


def test_add_assets(chart):
    chart.add_asset("cash")
    assert chart.chart.assets == ["cash"]


def test_add_capital(chart):
    chart.add_capital("equity")
    assert chart.chart.equity == ["equity"]


def test_add_liabilities(chart):
    chart.add_liability("ap")
    assert chart.chart.liabilities == ["ap"]


def test_add_income(chart):
    chart.add_income("sales")
    assert chart.chart.income == ["sales"]


def test_add_expenses(chart):
    chart.add_expense("cogs")
    assert chart.chart.expenses == ["cogs"]


def test_add_on_nonexistent_attribute():
    with pytest.raises(AttributeError):
        ChartCommand(chart=Chart())._add("no such attribute", [])


def test_split_on_caps():
    assert ContraIncome([], []).split_on_caps() == "Contra Income"



"""Chart of accounts data structure."""

from collections import Counter
from typing import Dict, List, Type

from pydantic import BaseModel  # type: ignore

from abacus.engine.accounts import (
    IncomeSummaryAccount,
    NullAccount,
    RegularAccount,
    RetainedEarnings,
    mapping,
)
from abacus.engine.base import AbacusError, AccountName, Amount, Entry, Pair


def repeated_names(xs):
    return [k for k, v in Counter(xs).items() if v > 1]


class Chart(BaseModel):
    """Chart of accounts."""

    assets: List[AccountName] = []
    expenses: List[AccountName] = []
    equity: List[AccountName] = []
    liabilities: List[AccountName] = []
    income: List[AccountName] = []
    retained_earnings_account: AccountName = "re"
    income_summary_account: AccountName = "current_profit"
    null_account = (
        "null"  # corresponding account to add starting balances using signle entries
    )
    contra_accounts: Dict[AccountName, List[AccountName]] = {}
    names: Dict[AccountName, str] = {"re": "Retained earnings"}
    codes: Dict[AccountName, str] = {}
    operations: Dict[str, Pair] = {}

    def make_entries(self, operation_names, amounts):
        return [
            Entry(*self.operations[on], amount)
            for on, amount in zip(operation_names, amounts)
        ]

    def five_types_of_accounts(self):
        return ("assets", "equity", "liabilities", "income", "expenses")

    def add_operation(self, name: str, debit: AccountName, credit: AccountName):
        self.operations[name] = (debit, credit)
        return self

    def set_code(self, account_name: AccountName, code: str):
        self.codes[account_name] = code
        return self

    def set_name(self, account_name: AccountName, title: str):
        self.names[account_name] = title
        return self

    def get_name(self, account_name: AccountName) -> str:
        try:
            return self.names[account_name]
        except KeyError:
            return account_name.replace("_", " ").strip().capitalize()

    def compose_name(self, account_name: AccountName):
        """Produce name like 'cash (Cash)'."""
        return account_name + " (" + self.get_name(account_name) + ")"

    def compose_name_long(self, account_name: AccountName):
        """Produce name like 'asset:cash (Cash)'."""
        t = self.get_account_type(account_name).__name__
        n = self.get_name(account_name)
        return t + ":" + account_name + " (" + n + ")"

    @property
    def duplicates(self):
        return repeated_names(self.account_names_all())

    def validate(self):
        if self.duplicates:
            raise AbacusError(
                f"Following account names were duplicated: {self.duplicates}"
            )
        return self

    def offset(
        self,
        account_name: AccountName,
        contra_account_names: AccountName | List[AccountName],
    ):
        """Add contra accounts to chart."""
        if account_name not in self.account_names_all():
            raise AbacusError(
                f"{account_name} must be specified in chart before adding {contra_account_names}."
            )
        if not isinstance(contra_account_names, list):
            contra_account_names = [contra_account_names]
        try:
            self.contra_accounts[account_name].extend(contra_account_names)
        except KeyError:
            self.contra_accounts[account_name] = contra_account_names
        return self

    def account_names_regular(self) -> List[AccountName]:
        return [
            account_name
            for attribute, _ in mapping()
            for account_name in getattr(self, attribute)
        ]

    def account_names_contra(self) -> List[AccountName]:
        return [v for values in self.contra_accounts.values() for v in values]

    def account_names_all(self) -> List[AccountName]:
        """Names of all regular accounts, contra accounts and unique accounts."""
        return (
            self.account_names_regular()
            + self.account_names_contra()
            + [
                self.retained_earnings_account,
                self.income_summary_account,
                self.null_account,
            ]
        )

    def account_names(self, cls: Type[RegularAccount]) -> List[AccountName]:
        """Return account names for account class *cls*."""
        reverse_dict = {
            Class.__name__: attribute for attribute, (Class, _) in mapping()
        }
        attribute = reverse_dict[cls.__name__]  # "assets" for example
        return getattr(self, attribute)

    def contra_account_pairs(self, cls: Type[RegularAccount]):
        """Yield pairs of regular and contra account names for given regular account type *cls*.
        This action unpacks self.conta_accounts to a sequence of tuples, each tuple is
        a pair of account names."""
        for regular_account_name, contra_account_names in self.contra_accounts.items():
            if regular_account_name in self.account_names(cls):
                for contra_account_name in contra_account_names:
                    yield regular_account_name, contra_account_name

    def get_account_type(self, account_name: AccountName) -> Type:
        return dict(yield_all_accounts(self))[account_name]

    def ledger_dict(self):
        """Create ledger dictionary from chart. Used to create Ledger class."""
        return {account_name: cls() for account_name, cls in yield_all_accounts(self)}

    def ledger(self, starting_balances: Dict[AccountName, Amount] = {}):
        """Create ledger from chart."""
        from abacus.engine.ledger import Ledger, to_multiple_entry

        ledger = Ledger.new(self)
        entries = to_multiple_entry(ledger, starting_balances).entries(
            self.null_account
        )
        ledger.post_many(entries)
        if (x := ledger[self.null_account].balance()) != 0:
            raise AbacusError(
                f"Balance of null account after adding starting balances must be 0, got {x}."
            )
        return ledger


def yield_all_accounts(chart: Chart):
    for attribute, (Class, ContraClass) in mapping():
        for account_name in getattr(chart, attribute):
            # regular accounts
            yield account_name, Class
            try:
                # contra accounts if found
                for contra_account_name in chart.contra_accounts[account_name]:
                    yield contra_account_name, ContraClass
            except KeyError:
                pass
    yield chart.retained_earnings_account, RetainedEarnings
    yield chart.income_summary_account, IncomeSummaryAccount
    yield chart.null_account, NullAccount



from abacus.engine.report import BalanceSheet, IncomeStatement


def test_printing():
    bs = BalanceSheet(
        assets=({"cash": 1130, "receivables": 25, "goods_for_sale": 45}),
        capital=({"equity": 1000, "re": 75}),
        liabilities=({"payables": 50, "divp": 75}),
    )
    inc = IncomeStatement(
        income=({"sales": 760}),
        expenses=({"cogs": 440, "sga": 400}),
    )
    rename_dict = {
        "re": "Retained earnings",
        "divp": "Dividend due",
        "cogs": "Cost of goods sold",
        "sga": "Selling, general and adm. expenses",
    }
    assert bs.view(rename_dict)
    bs.print_rich(rename_dict)
    assert inc.view(rename_dict)
    inc.print_rich(rename_dict)



# This is pytest unit test file for experimental/entries.py

import pytest  # type: ignore

from abacus.engine.base import Entry
from abacus.engine.entries import CsvFile


@pytest.fixture
def path(tmp_path):
    return tmp_path / "entries.csv"


def test_yield_entries(path):
    file = CsvFile(path)
    file.touch()
    e1, e2 = Entry("cash", "equity", 499), Entry("cash", "equity", 501)
    file.append(e1)
    file.append(e2)
    assert list(file.yield_entries()) == [e1, e2]



from abacus import BalanceSheet, Chart, Entry, IncomeStatement
from abacus.engine.closing import ClosingEntries, make_closing_entries

# Create chart of accounts
chart = Chart(
    assets=["cash"],
    equity=["equity"],
    expenses=["salaries", "rent"],
    liabilities=[],
    income=["services"],
).offset("services", ["cashback"])


# Account balances are known from previous period end
starting_balances = {"cash": 1400, "equity": 1500, "re": -100}

# Create general ledger and post new entries
ledger = (
    chart.ledger(starting_balances)
    .post(Entry(debit="rent", credit="cash", amount=200))
    .post(Entry(debit="cash", credit="services", amount=825))
    .post(Entry(debit="cashback", credit="cash", amount=25))
    .post(Entry(debit="salaries", credit="cash", amount=400))
)


def test_closing_entries():
    assert make_closing_entries(chart, ledger) == ClosingEntries(
        closing_contra_income=[Entry(debit="services", credit="cashback", amount=25)],
        closing_contra_expense=[],
        closing_income=[Entry(debit="services", credit="current_profit", amount=800)],
        closing_expense=[
            Entry(debit="current_profit", credit="salaries", amount=400),
            Entry(debit="current_profit", credit="rent", amount=200),
        ],
        closing_isa=Entry(debit="current_profit", credit="re", amount=200),
    )


def test_closing_entries_for_is():
    assert ledger.closing_entries_for_income_statement(chart) == [
        Entry(debit="services", credit="cashback", amount=25)
    ]


def test_closing_entries_for_bs():
    assert ledger.closing_entries_for_balance_sheet(chart) == [
        Entry(debit="services", credit="cashback", amount=25),
        Entry(debit="services", credit="current_profit", amount=800),
        Entry(debit="current_profit", credit="salaries", amount=400),
        Entry(debit="current_profit", credit="rent", amount=200),
        Entry(debit="current_profit", credit="re", amount=200),
    ]


def test_balance_sheet():
    assert ledger.balance_sheet(chart) == BalanceSheet(
        assets={"cash": 1600}, capital={"equity": 1500, "re": 100}, liabilities={}
    )


def test_balances():
    assert ledger.condense().close(chart).balances() == {
        "current_profit": 0,
        "null": 0,
        "cashback": 0,
        "cash": 1600,
        "equity": 1500,
        "re": 100,
        "rent": 0,
        "salaries": 0,
        "services": 0,
    }


def test_current_balances_n():
    assert ledger.income_statement(chart).current_profit() == 200


def test_income_statement():
    assert ledger.income_statement(chart) == IncomeStatement(
        income={"services": 800}, expenses={"salaries": 400, "rent": 200}
    )



"""
Usage:
  bx chart init
  bx chart add --asset     <account_name> [(--title <title>)] [(--code <code>)] 
  bx chart add --capital   <account_name> [(--title <title>)] [(--code <code>)] 
  bx chart add --liability <account_name> [(--title <title>)] [(--code <code>)] 
  bx chart add --income    <account_name> [(--title <title>)] [(--code <code>)] 
  bx chart add --expense   <account_name> [(--title <title>)] [(--code <code>)] 
  bx chart offset <account_name> <contra_account_names>...
  bx chart set <account_name> [(--title <title>)] [(--code <code>)]
  bx chart set --retained-earnings <account_name> [(--title <title>)] [(--code <code>)]
  bx chart set --income-summary-account <account_name> [(--title <title>)] [(--code <code>)]
  bx chart set --null-account <account_name> [(--title <title>)] [(--code <code>)]
  bx chart alias (--operation <name>) (--debit <debit>) (--credit <credit>) [(--requires <requires>)]
  bx chart show [--json]
  bx chart erase --force
  bx ledger init [<file>]
  bx ledger post <debit> <credit> <amount> [--title <title>]
  bx ledger post (--debit <debit>) (--credit <credit>) (--amount <amount>) [(--title <title>)]
  bx ledger post --operation (<operations> <amounts>)...  [(--title <title>)]
  bx ledger close
  bx ledger show
  bx ledger erase --force
  bx report --trial-balance
  bx report --income-statement [--json | --rich]
  bx report --balance-sheet [--json | --rich]
  bx account <account_name>
  bx account <account_name> --assert <value>
  bx balances [--nonzero]
"""
import json
import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List

from docopt import docopt  # type: ignore

from abacus.engine.accounts import CreditAccount, DebitAccount
from abacus.engine.base import AbacusError, AccountName, Amount, Entry
from abacus.engine.chart import Chart
from abacus.engine.closing import make_closing_entries
from abacus.engine.entries import CsvFile
from abacus.engine.ledger import Ledger, to_multiple_entry


def cwd() -> Path:
    return Path(os.getcwd())


def get_chart_path() -> Path:
    return cwd() / "chart.json"


def get_entries_path() -> Path:
    return cwd() / "entries.csv"


def read_json(path):
    return json.loads(Path(path).read_text(encoding="utf-8"))


def init_chart(path: Path):
    """Write empty chart to *path* if does not file exist."""
    if not path.exists():
        ChartCommand(chart=Chart()).write(path)
    else:
        raise AbacusError


def print_chart(chart: Chart):
    def name(account_name):
        return chart.compose_name(account_name)

    print("Accounts:")
    for attribute in chart.five_types_of_accounts():
        account_names = getattr(chart, attribute)
        if account_names:
            print(attribute.capitalize() + ":", ", ".join(map(name, account_names)))
    if chart.contra_accounts:
        print("Contra accounts:")
        for key, names in chart.contra_accounts.items():
            print("  -", contra_phrase(name(key), map(name, names)))
    print_re(chart)
    if chart.operations:
        print("Operation aliases:")
        for key, (debit, credit) in chart.operations.items():
            print("  -", key, f"(debit is {debit}, credit is {credit})")


def print_re(chart):
    print(
        "Retained earnings account:",
        chart.compose_name(chart.retained_earnings_account) + ".",
    )


def contra_phrase(account_name, contra_account_names):
    return account_name + " is offset by " + ", ".join(contra_account_names)


@dataclass
class ChartCommand:
    chart: Chart

    @classmethod
    def read(cls, path: Path):
        return ChartCommand(chart=Chart.parse_file(path))

    def write(self, path: Path):
        content = self.chart.json(indent=4, ensure_ascii=True)
        Path(path).write_text(content, encoding="utf-8")

    def set_retained_earnings(self, account_name):
        """Set name of null account (override default name).

        Command:
        ```
        bx set --retained-earnings <account_name>
        ```
        """
        self.chart.retained_earnings_account = account_name
        return self

    def set_null_account(self, account_name):
        """Set name of null account (override default name).

        Command:
        ```
        bx set --null-account <account_name>
        ```
        """
        self.chart.null_account = account_name
        return self

    def set_isa(self, account_name):
        """Set name of income summary account (override default name).

        Command:
        ```
        bx set --income-summary-account <account_name>
        ```
        """
        self.chart.income_summary_account = account_name
        return self

    def _add(self, attribute: str, account_name: str):
        setattr(self.chart, attribute, getattr(self.chart, attribute) + [account_name])

    def add_asset(self, account_name: str):
        self._add("assets", account_name)

    def add_capital(self, account_name: str):
        # adding account_names to 'equity' attribute
        self._add("equity", account_name)

    def add_liability(self, account_name: str):
        self._add("liabilities", account_name)

    def add_income(self, account_name: str):
        self._add("income", account_name)

    def add_expense(self, account_name: str):
        self._add("expenses", account_name)

    def offset(self, account_name, contra_account_names) -> str:
        for contra_account_name in contra_account_names:
            self.chart.offset(account_name, contra_account_name)
        if len(contra_account_names) > 1:
            text = "Added contra accounts: "
        else:
            text = "Added contra account: "
        return text + contra_phrase(account_name, contra_account_names) + "."

    def alias(self, name: str, debit: AccountName, credit: AccountName):
        self.chart.add_operation(name, debit, credit)
        return f"Added operation: {name} (debit is {debit}, credit is {credit})."

    def set_name(self, account_name, title) -> str:
        self.chart.set_name(account_name, title)
        return "Added account title: " + self.chart.compose_name(account_name) + "."

    def set_code(self, account_name, code: str) -> str:
        self.chart.set_code(account_name, code)
        return f"Set code {code} for account {account_name}."


def chart_command(arguments: Dict, chart_path: Path):
    if arguments["init"]:
        try:
            init_chart(chart_path)
            print("Created", chart_path)
            sys.exit(0)
        except AbacusError:
            sys.exit(f"File already exists: {chart_path}")
    if arguments["erase"] and arguments["--force"]:
        if chart_path.exists():
            chart_path.unlink(missing_ok=True)
            print("Deleted", chart_path)
        else:
            print("File not found:", chart_path)
        sys.exit(0)
    # load chart file
    try:
        holder = ChartCommand.read(chart_path)
    except FileNotFoundError:
        sys.exit(
            f"File not found (expected {chart_path}).\nUse `bx chart init` to create chart file."
        )
    # commands below assume chart file exists
    account_name = arguments["<account_name>"]
    if arguments["set"] or arguments["add"]:
        if arguments["--title"]:
            title = arguments["<title>"]
            msg = holder.set_name(account_name, title)
            print(msg)
            holder.write(chart_path)
        if arguments["--code"]:
            code = arguments["<code>"]
            if not code:
                raise ValueError(f"Code must be provided, got {code} in {arguments}.")
            msg = holder.set_code(account_name, code)
            print(msg)
            holder.write(chart_path)
        # set
        if arguments["--retained-earnings"]:
            holder.set_retained_earnings(account_name)
            print_re(holder.chart)
        elif arguments["--income-summary-account"]:
            holder.set_isa(account_name)
        elif arguments["--null-account"]:
            holder.set_null_account(account_name)
        # add
        if arguments["--asset"]:
            holder.add_asset(account_name)
        elif arguments["--capital"]:
            holder.add_capital(account_name)
        elif arguments["--liability"]:
            holder.add_liability(account_name)
        elif arguments["--income"]:
            holder.add_income(account_name)
        elif arguments["--expense"]:
            holder.add_expense(account_name)
        holder.write(chart_path)
        print("Added account to chart:", account_name + ".")
    elif arguments["offset"]:
        msg = holder.offset(
            arguments["<account_name>"], arguments["<contra_account_names>"]
        )
        print(msg)
        holder.write(chart_path)
    elif arguments["alias"]:
        msg = holder.alias(
            arguments["<name>"], arguments["<debit>"], arguments["<credit>"]
        )
        holder.write(chart_path)
        print(msg)
    elif arguments["show"]:
        if arguments["--json"]:
            print(holder.chart.json())
        else:
            print_chart(holder.chart)


@dataclass
class LedgerCommand:
    chart: Chart
    csv_file: CsvFile

    def add_starting_balances(self, starting_balances_dict: Dict):
        self.chart.ledger(starting_balances_dict)  # check that all accounts are present
        me = to_multiple_entry(self.chart.ledger(), starting_balances_dict)
        entries = me.entries(self.chart.null_account)
        self.csv_file.append_many(entries)
        for entry in entries:
            print("Posted entry:", entry)

    def add_operations(self, operations: List[str], amounts: List[str]):
        entries = self.chart.make_entries(operations, amounts)
        self.csv_file.append_many(entries)

    def add_closing_entries(self):
        ledger = Ledger.new(self.chart)
        ledger.post_many(entries=self.csv_file.yield_entries())
        closing_entries = make_closing_entries(self.chart, ledger).all()
        self.csv_file.append_many(entries=closing_entries)
        print("Added closing entries:", closing_entries)

    def append_entry(self, debit: str, credit: str, amount: str):
        entry = Entry(debit=debit, credit=credit, amount=Amount(amount))
        self.csv_file.append(entry)
        print("Posted entry:", entry)


def ledger_command(arguments: Dict, entries_path: Path, chart_path: Path):
    file = CsvFile(entries_path)
    if arguments["erase"] and arguments["--force"]:
        file.erase()
        sys.exit(0)
    chart = Chart.parse_file(chart_path)
    holder = LedgerCommand(chart=chart, csv_file=file)
    if arguments["init"]:
        file.touch()
        print("Created", entries_path)
        if arguments["<file>"]:
            balances = read_starting_balances(arguments["<file>"])
            print("Starting balances:", balances)
            holder.add_starting_balances(balances)
    elif arguments["post"]:
        if arguments["--operation"]:
            holder.add_operations(arguments["<operations>"], arguments["<amounts>"])
        else:
            holder.append_entry(
                debit=arguments["<debit>"],
                credit=arguments["<credit>"],
                amount=arguments["<amount>"],
            )
    elif arguments["close"]:
        holder.add_closing_entries()
    elif arguments["show"]:
        for entry in file.yield_entries():
            print(entry.debit, entry.credit, entry.amount, sep="\t")


def touches_isa(chart: Chart, entry: Entry) -> bool:
    """True if entry touches income summary account."""
    isa = chart.income_summary_account
    return (entry.debit == isa) or (entry.credit == isa)


def filter_for_income_statement(entries: List[Entry], chart: Chart):
    """Filter entries that do not close income accounts."""

    def not_touches_isa(entry):
        return not touches_isa(chart, entry)

    return filter(not_touches_isa, entries)


def report_command(arguments: Dict, entries_path: Path, chart_path: Path):
    chart = Chart.parse_file(chart_path)
    entries = CsvFile(entries_path).yield_entries()
    ledger = Ledger.new(chart)
    if arguments["--trial-balance"]:
        ledger.post_many(entries)
        print(ledger.trial_balance(chart))
    elif arguments["--balance-sheet"]:
        ledger.post_many(entries)
        statement = ledger.balance_sheet(chart)
        if arguments["--json"]:
            print(statement.json())
        elif arguments["--rich"]:
            statement.print_rich(chart.names)
        else:
            print("Balance sheet")
            print(statement.view(chart.names))
    elif arguments["--income-statement"]:
        entries = filter_for_income_statement(entries, chart)
        ledger.post_many(entries)
        statement = ledger.income_statement(chart)
        if arguments["--json"]:
            print(statement.json())
        elif arguments["--rich"]:
            statement.print_rich(chart.names)
        else:
            print("Income statement")
            print(statement.view(chart.names))


def side(ledger, account_name):
    match ledger[account_name]:
        case DebitAccount(_, _):
            return "debit"
        case CreditAccount(_, _):
            return "credit"
        case _:
            raise TypeError(ledger[account_name].__class__.__name__)


def print_account_info(ledger, chart, account_name: str):
    account = ledger[account_name]
    print("     Account:", chart.get_name(account_name))
    print("  Short name:", account_name)
    print("   Long name:", chart.compose_name_long(account_name))
    print("        Code:", chart.codes.get(account_name, "not assigned"))
    print("        Type:", account.split_on_caps())
    print("      Debits:", str(account.debits) + ",", "total", sum(account.debits))
    print("     Credits:", str(account.credits) + ",", "total", sum(account.credits))
    print("     Balance:", account.balance())
    print("Balance side:", side(ledger, account_name).capitalize())


def assert_command(arguments, entries_path, chart_path):
    chart = Chart.parse_file(chart_path)
    entries = CsvFile(entries_path).yield_entries()
    ledger = Ledger.new(chart).post_many(entries)
    assert_account_balance(ledger, arguments["<account_name>"], arguments["<balance>"])


def assert_account_balance(ledger, account_name: str, assert_amount: str):
    amount = ledger[account_name].balance()
    expected_amount = Amount(assert_amount)
    if amount != expected_amount:
        sys.exit(
            f"Check failed for {account_name}, expected balance is {expected_amount}, got {amount}."
        )
    print(
        f"Check passed for account {account_name}, account balance is {expected_amount} as expected."
    )


def account_command(arguments, entries_path, chart_path):
    """
    bx account <account_name>
    bx account <account_name> --balance <value>
    """
    chart = Chart.parse_file(chart_path)
    ledger = process_full_ledger(chart_path, entries_path)
    account_name = arguments["<account_name>"]
    if v := arguments["<value>"]:
        assert_account_balance(ledger, account_name, v)
    else:
        print_account_info(ledger, chart, account_name)


def process_full_ledger(chart_path: Path, entries_path: Path) -> Ledger:
    chart = Chart.parse_file(chart_path)
    entries = CsvFile(entries_path).yield_entries()
    return Ledger.new(chart).post_many(entries)


def show_balances_command(arguments, entries_path, chart_path):
    """
    bx balances [--nonzero]
    """
    # this command always returns a json, there is no --json flag
    ledger = process_full_ledger(chart_path, entries_path)
    balances = ledger.balances()
    if arguments["--nonzero"]:
        balances = ledger.nonzero_balances()
    print(json.dumps(balances))


def main():
    entries_path = get_entries_path()
    chart_path = get_chart_path()
    arguments = docopt(__doc__, version="0.5.1")
    if arguments["chart"]:
        chart_command(arguments, chart_path)
    elif arguments["ledger"]:
        ledger_command(arguments, entries_path, chart_path)
    elif arguments["report"]:
        report_command(arguments, entries_path, chart_path)
    elif arguments["account"]:
        account_command(arguments, entries_path, chart_path)
    elif arguments["balances"]:
        show_balances_command(arguments, entries_path, chart_path)
    else:
        sys.exit("Command not recognized. Use bx --help for reference.")


def read_starting_balances(path: str) -> Dict:
    """Read starting balances from file *path*."""
    return json.loads(Path(path).read_text(encoding="utf-8"))


if __name__ == "__main__":
    main()



import json
import pathlib
import subprocess

import pytest  # type: ignore


@pytest.fixture
def commands():
    path = pathlib.Path(__file__).parent / "test_cli.sh"
    return path.read_text().strip().split("\n")


def test_n_commands(commands):
    assert len(commands) == 17


def test_runner(tmpdir):
    result = subprocess.run(["python", "--version"], cwd=tmpdir)
    assert result.returncode == 0


def test_run_bx_help(tmpdir):
    result = subprocess.run(
        "python -m abacus.bx --help".split(), shell=True, cwd=tmpdir
    )
    assert result.returncode == 0


def test_few_commands(commands, tmpdir):
    for line in commands:
        command = line
        result = subprocess.run(
            command, shell=True, capture_output=True, text=True, cwd=tmpdir
        )
        assert result.returncode == 0
    # final check
    assert json.loads(result.stdout) == {
        "cash": 200,
        "goods": 800,
        "equity": 1000,
    }



# I need to save debit(string), credit(string), amount(float) tuples to a CSV file
# The CSV file should be open for appending.

import csv
from dataclasses import dataclass
from pathlib import Path
from typing import List

from abacus.engine.base import Amount, Entry


@dataclass
class CsvFile:
    """File that stores accounting entries in CSV format."""

    path: Path

    def append(self, entry: Entry) -> None:
        with open(self.path, "a", newline="", encoding="utf-8") as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow([entry.debit, entry.credit, entry.amount])

    def append_many(self, entries: List[Entry]) -> None:
        with open(self.path, "a", newline="", encoding="utf-8") as csvfile:
            writer = csv.writer(csvfile)
            for entry in entries:
                writer.writerow([entry.debit, entry.credit, entry.amount])

    def yield_entries(self):
        with open(self.path, "r", newline="", encoding="utf-8") as csvfile:
            reader = csv.reader(csvfile)
            for row in reader:
                yield Entry(row[0], row[1], Amount(row[2]))

    def touch(self):
        if not self.path.exists():
            self.path.touch()

    def erase(self):
        if self.path.exists():
            self.path.unlink()



"""Classes to different types of accounts, all based on TAccount."""
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List

from abacus.engine.base import AccountName, Amount, Entry


@dataclass
class TAccount(ABC):
    """Parent class for T-account with debits and credits."""

    debits: List[Amount] = field(default_factory=list)
    credits: List[Amount] = field(default_factory=list)

    @abstractmethod
    def balance(self) -> Amount:
        """Return account balance."""

    def debit(self, amount: Amount):
        """Append *amount* to debit side of the account."""
        self.debits.append(amount)
        return self

    def credit(self, amount: Amount):
        """Append *amount* to credit side of the account."""
        self.credits.append(amount)
        return self

    def deep_copy(self):
        """Replicate this account as new data structure."""
        return self.__class__(self.debits.copy(), self.credits.copy())

    def condense(self):
        """Create a new account of the same type with only one value as account balance."""
        return self.empty().topup(self.balance())

    def empty(self):
        """Create a new empty account of the same type."""
        return self.__class__()

    def topup(self, balance):
        """Add balance to a proper side of account."""
        match self:
            case DebitAccount(_, _):
                return self.debit(balance)
            case CreditAccount(_, _):
                return self.credit(balance)

    def split_on_caps(self) -> str:
        return " ".join(re.findall("[A-Z][^A-Z]*", self.__class__.__name__))


class DebitAccount(TAccount):
    def balance(self) -> Amount:
        return sum(self.debits) - sum(self.credits)


class CreditAccount(TAccount):
    def balance(self) -> Amount:
        return sum(self.credits) - sum(self.debits)


class Permanent:
    pass


class Temporary:
    pass


class RegularAccount(Permanent):
    pass


class Asset(DebitAccount, RegularAccount):
    pass


class Expense(DebitAccount, RegularAccount):
    pass


class Capital(CreditAccount, RegularAccount):
    pass


class Liability(CreditAccount, RegularAccount):
    pass


class Income(CreditAccount, RegularAccount):
    pass


class Unique:
    """There should be just one account of this class in any ledger.
    `IncomeSummaryAccount`,`RetainedEarnings` and `NullAccount` are unique.
    """


class Transferable(TAccount):
    def transfer(self, my_name: AccountName, dest_name: AccountName) -> Entry:
        """Make entry to transfer balances from account of named *my_name* to *dest_name* account.
        The entry will differ depending if *my_name* is debit or credit account."""
        amount = self.balance()
        match self:
            case DebitAccount(_, _):
                return Entry(dest_name, my_name, amount)
            case CreditAccount(_, _):
                return Entry(my_name, dest_name, amount)
            case _:
                raise TypeError


# Can equally be CreditAccount or DebitAccount, making it a ContraAccount
# keeps null at the bottom of trail balance, so it is less distracting.
class NullAccount(CreditAccount, Unique, Temporary):
    pass


class IncomeSummaryAccount(CreditAccount, Unique, Transferable, Temporary):
    pass


class RetainedEarnings(Capital, Unique):
    pass


class ContraAccount(Transferable):
    pass


class ContraAsset(CreditAccount, ContraAccount, Permanent):
    pass


class ContraExpense(CreditAccount, ContraAccount, Temporary):
    pass


class ContraCapital(DebitAccount, ContraAccount, Permanent):
    pass


class ContraLiability(DebitAccount, ContraAccount, Permanent):
    pass


class ContraIncome(DebitAccount, ContraAccount, Temporary):
    pass


def mapping():
    """Relate Chart class attributes (keys) to actual classes (values)."""
    return dict(
        assets=(Asset, ContraAsset),
        expenses=(Expense, ContraExpense),
        equity=(Capital, ContraCapital),
        liabilities=(Liability, ContraLiability),
        income=(Income, ContraIncome),
    ).items()



"""
A minimal, yet valid double-entry accounting system in Python that can:

- define a chart of accounts and create a ledger,
- post entries to ledger,
- produce trial balance,
- post adjustment entries,
- close accounts at accounting period end,
- produce income statement and balance sheet,
- make post-close entries.

Simplifying assumptions:

- one currency
- no journals, entries posted directly to ledger
- entry holds just amount, debit and credit accounts, no title or date
- only balance sheet and income statement reports, no cash flow
- no non-negativity checks (can distribute dividend greater than profit)
- balances stick to one side (debit or credit) based on type of account

"""
from .engine.base import AbacusError, Amount, Entry  # noqa: F401
from .engine.chart import Chart  # noqa: F401
from .engine.ledger import Ledger  # noqa: F401
from .engine.report import BalanceSheet, IncomeStatement  # noqa: F401



"""Baisc data types used across package."""

from dataclasses import dataclass
from typing import Tuple

AccountName = str
Amount = int  # Can be changed to Decimal


@dataclass
class Entry:
    """Pure accounting entry (without title, date, etc.)."""

    debit: AccountName
    credit: AccountName
    amount: Amount


class AbacusError(Exception):
    """Custom error class."""


"""Pair of debit and credit account names."""
Pair = Tuple[AccountName, AccountName]


@dataclass
class NamedEntry:
    operation: str
    amount: Amount

    def to_entry(self, operation_dict: dict[str, Pair]) -> Entry:
        dr, cr = operation_dict[self.operation]
        return Entry(dr, cr, self.amount)


def sum_second(xs):
    return sum(x for _, x in xs)


@dataclass
class MultipleEntry:
    debit_entries: list[tuple[AccountName, Amount]]
    credit_entries: list[tuple[AccountName, Amount]]

    def __post_init__(self):
        self.validate()

    def validate(self):
        if sum_second(self.debit_entries) != sum_second(self.credit_entries):
            raise AbacusError(["Invalid multiple entry", self])
        return self

    def entries(self, null_account_name: AccountName):
        return [
            Entry(account_name, null_account_name, amount)
            for (account_name, amount) in self.debit_entries
        ] + [
            Entry(null_account_name, account_name, amount)
            for (account_name, amount) in self.credit_entries
        ]



"""Ledger data structure, can be created from chart, used to post entries and produce reports."""
from collections import UserDict
from typing import Dict, List, Tuple, Type

from abacus.engine.accounts import CreditAccount, DebitAccount, TAccount
from abacus.engine.base import AbacusError, AccountName, Amount, Entry, MultipleEntry
from abacus.engine.chart import Chart


class Ledger(UserDict[AccountName, TAccount]):
    """General ledger that holds all accounts. Accounts are referenced by name."""

    @classmethod
    def new(cls, chart: Chart) -> "Ledger":
        """Create an empty ledger from chart."""
        return Ledger(chart.ledger_dict())

    def apply(self, attribute: str):
        """Apply an *attribute* to all accounts in ledger and return a new ledger."""
        return Ledger(
            {
                account_name: getattr(taccount, attribute)()
                for account_name, taccount in self.items()
            }
        )

    def deep_copy(self) -> "Ledger":
        """Return a copy of the ledger, taking care of copying
        debit and credit lists in accounts."""
        return self.apply("deep_copy")

    def condense(self) -> "Ledger":
        """Purge data from ledger leaving only account balances.
        Creates a copy of ledger."""
        return self.apply("condense")

    def empty(self) -> "Ledger":
        """Purge data from ledger. Creates a copy of ledger."""
        return self.apply("empty")

    def balances(self) -> Dict[AccountName, Amount]:
        """Return account balances."""
        return dict(self.apply("balance").data)

    def nonzero_balances(self) -> Dict[AccountName, Amount]:
        """Return nonzero account balances."""
        return {k: v for k, v in self.balances().items() if v != 0}

    def subset(self, classes: Type | List[Type]) -> "Ledger":
        """Filter ledger by account type."""
        if not isinstance(classes, list):
            classes = [classes]

        def belongs(t_account: TAccount):
            return any(isinstance(t_account, cls) for cls in classes)

        return Ledger(
            [
                (account_name, t_account)
                for account_name, t_account in self.items()
                if belongs(t_account)
            ]
        )

    def post(self, entry: Entry) -> "Ledger":
        # Note that post() mutates the existing data structure
        post_entry(self, entry)
        return self

    def post_many(self, entries: List[Entry]) -> "Ledger":
        post_entries(self, entries)
        return self

    def closing_entries_for_income_statement(self, chart: Chart):
        from abacus.engine.closing import close_contra_expense as ce
        from abacus.engine.closing import close_contra_income as ci

        return ci(chart, self) + ce(chart, self)

    def close_some(self, chart: Chart) -> "Ledger":
        """Close contra accounts corresponding to income and expense.
        Returns a copy of a ledger.
        """
        closing_entries = self.closing_entries_for_income_statement(chart)
        return self.condense().post_many(closing_entries)

    def income_statement(self, chart: Chart, post_close_entries=None):
        from abacus.engine.report import IncomeStatement

        if post_close_entries is None:
            post_close_entries = []
        ledger = self.condense().close_some(chart).post_many(post_close_entries)
        return IncomeStatement.new(ledger)

    def closing_entries_for_balance_sheet(self, chart: Chart):
        from abacus.engine.closing import flush_permanent_accounts, make_closing_entries

        closing_entries_struct = make_closing_entries(chart, self)
        return closing_entries_struct.all() + flush_permanent_accounts(chart, self)

    def close(self, chart: Chart) -> "Ledger":
        """Close all accounts related to retained earnings and to permanent accounts.
        Returns a copy of a ledger.
        """
        closing_entries = self.closing_entries_for_balance_sheet(chart)
        return self.condense().post_many(closing_entries)

    def balance_sheet(self, chart: Chart, post_close_entries=None):
        from abacus.engine.report import BalanceSheet

        if post_close_entries is None:
            post_close_entries = []
        ledger = self.close(chart).post_many(post_close_entries)
        return BalanceSheet.new(ledger)

    def trial_balance(self, chart: Chart):
        from abacus.engine.report import view_trial_balance

        return view_trial_balance(chart, self)


def post_entry(ledger: Ledger, entry: Entry) -> None:
    ledger[entry.debit].debit(entry.amount)
    ledger[entry.credit].credit(entry.amount)


def post_entries(ledger: Ledger, entries: List[Entry]):
    ledger, failed_entries = unsafe_post_entries(ledger, entries)
    if failed_entries:
        raise AbacusError(failed_entries)


def unsafe_post_entries(
    ledger: Ledger, entries: List[Entry]
) -> Tuple[Ledger, List[Entry]]:
    failed = []
    for entry in entries:
        try:
            post_entry(ledger, entry)
        except KeyError:
            failed.append(entry)
    return ledger, failed


def to_multiple_entry(ledger: Ledger, starting_balances: dict) -> MultipleEntry:
    debit_entries = []
    credit_entries = []
    for account_name in starting_balances.keys():
        try:
            ledger[account_name]
        except KeyError:
            raise AbacusError(f"Account {account_name} does not exist in ledger.")
    for account_name, amount in starting_balances.items():
        match ledger[account_name]:
            case DebitAccount(_, _):
                debit_entries.append((account_name, amount))
            case CreditAccount(_, _):
                credit_entries.append((account_name, amount))
    return MultipleEntry(debit_entries, credit_entries).validate()



from abacus.engine.chart import Chart
from abacus.engine.report import BalanceSheet, IncomeStatement

chart = Chart(
    assets=["cash", "receivables", "goods_for_sale"],
    expenses=["cogs", "sga"],
    equity=["equity"],
    liabilities=["divp", "payables"],
    income=["sales"],
)
chart.operations = operations = dict(
    pay_shareholder_capital=("cash", "equity"),
    buy_goods_for_cash=("goods_for_sale", "cash"),
    invoice_buyer=("receivables", "sales"),
    transfer_goods_sold=("cogs", "goods_for_sale"),
    accept_payment=("cash", "receivables"),
    accrue_salary=("sga", "payables"),
    pay_salary=("payables", "cash"),
)
named_entries = [
    # start a company
    ("pay_shareholder_capital", 501),
    ("pay_shareholder_capital", 499),
    # acquire goods
    ("buy_goods_for_cash", 820),
    # one order
    ("invoice_buyer", 600),
    ("transfer_goods_sold", 360),
    ("accept_payment", 549),
    # pay labor
    ("accrue_salary", 400),
    ("pay_salary", 345),
    # another order
    ("invoice_buyer", 160),
    ("transfer_goods_sold", 80),
    ("accept_payment", 80),
]


def fst(xs):
    return [x[0] for x in xs]


def scd(xs):
    return [x[1] for x in xs]


def test_named_entries_income_statement():
    entries = chart.make_entries(fst(named_entries), scd(named_entries))
    income_st = chart.ledger().post_many(entries).income_statement(chart)
    assert income_st == IncomeStatement(
        income={"sales": 760}, expenses={"cogs": 440, "sga": 400}
    )


def test_named_entries_balance_sheet():
    entries = chart.make_entries(fst(named_entries), scd(named_entries))
    assert chart.ledger().post_many(entries).balance_sheet(chart) == BalanceSheet(
        assets={"cash": 464, "receivables": 131, "goods_for_sale": 380},
        capital={"equity": 1000, "re": -80},
        liabilities={"divp": 0, "payables": 55},
    )



from abacus.engine.accounts import (
    Asset,
    Capital,
    ContraAsset,
    ContraCapital,
    ContraIncome,
    Expense,
    Income,
    IncomeSummaryAccount,
    NullAccount,
    RetainedEarnings,
)
from abacus.engine.base import Entry, MultipleEntry
from abacus.engine.chart import Chart
from abacus.engine.ledger import Ledger, to_multiple_entry, unsafe_post_entries


def test_ledger(ledger):
    assert {
        k: v
        for k, v in ledger.subset([Asset, Expense, ContraIncome])
        .deep_copy()
        .condense()
        .balances()
        .items()
        if v
    } == {"cash": 1000, "ar": 200, "refunds": 50}


def test_ledger_new():
    assert Ledger.new(Chart(assets=["cash"], equity=["equity"])).data == {
        "cash": Asset(debits=[], credits=[]),
        "equity": Capital(debits=[], credits=[]),
        "re": RetainedEarnings(debits=[], credits=[]),
        "current_profit": IncomeSummaryAccount(debits=[], credits=[]),
        "null": NullAccount(debits=[], credits=[]),
    }


def test_ledger_post():
    ledger1 = Ledger.new(Chart(assets=["cash"], equity=["equity"]))
    ledger1.post(Entry(debit="cash", credit="equity", amount=799))
    ledger1.post(Entry(debit="cash", credit="equity", amount=201))
    assert ledger1["cash"].debits == ledger1["equity"].credits == [799, 201]
    assert ledger1["cash"].balance() == ledger1["equity"].balance() == 1000


def test_ledger_deep_copy():
    ledger1 = Ledger.new(Chart(assets=["cash"], equity=["equity"]))
    ledger1.post(Entry(debit="cash", credit="equity", amount=799))
    ledger1.post(Entry(debit="cash", credit="equity", amount=201))
    ledger2 = ledger1.deep_copy()
    ledger2.post(Entry(debit="cash", credit="equity", amount=-1000))
    # ledger1 not affected
    assert ledger1["cash"].balance() == ledger1["equity"].balance() == 1000
    # ledger2 changed
    assert ledger2["cash"].balance() == ledger2["equity"].balance() == 0


def test_ledger_report_balance_sheet(chart, ledger):
    assert ledger.balance_sheet(chart).dict() == {
        "assets": {"cash": 1000, "ar": 200, "goods": 0},
        "capital": {"equity": 1000, "re": 200},
        "liabilities": {},
    }


def test_ledger_report_income_statement(chart, ledger):
    assert ledger.income_statement(chart).dict() == {
        "income": {"sales": 200},
        "expenses": {"cogs": 0, "sga": 0},
    }


def test_journal_with_starting_balance():
    chart = Chart(
        assets=["cash"],
        equity=["equity"],
        expenses=["salaries", "rent"],
        liabilities=[],
        income=["services"],
    )
    ledger = Ledger.new(chart)

    # Account balances are known from previous period end
    starting_balances = {"cash": 1400, "equity": 1500, "re": -100}
    assert to_multiple_entry(ledger, starting_balances) == MultipleEntry(
        [("cash", 1400)],
        [("equity", 1500), ("re", -100)],
    )


def test_post_many():
    chart = Chart(
        assets=["cash", "goods_for_sale"],
        expenses=["cogs", "sga"],
        equity=["equity"],
        income=["sales"],
        liabilities=[],
    ).offset("sales", ["discounts", "cashback"])

    starting_balances = {"cash": 10, "goods_for_sale": 10, "equity": 20}
    ledger = chart.ledger(starting_balances)
    e1 = Entry(debit="cash", credit="equity", amount=1000)  # pay in capital
    e2 = Entry(
        debit="goods_for_sale", credit="cash", amount=250
    )  # acquire goods worth 250
    e3 = Entry(
        credit="goods_for_sale", debit="cogs", amount=200
    )  # sell goods worth 200
    e4 = Entry(credit="sales", debit="cash", amount=400)  # for 400 in cash
    e5 = Entry(credit="cash", debit="sga", amount=50)  # administrative expenses
    ledger.post_many([e1, e2, e3, e4, e5])
    assert ledger.income_statement(chart).current_profit() == 150


def test_make_ledger():
    chart = Chart(
        assets=["cash"],
        equity=["equity"],
    )
    assert chart.ledger() == {
        "cash": Asset(debits=[], credits=[]),
        "equity": Capital(debits=[], credits=[]),
        "current_profit": IncomeSummaryAccount(debits=[], credits=[]),
        "re": RetainedEarnings(debits=[], credits=[]),
        "null": NullAccount(debits=[], credits=[]),
    }


def test_unsafe_process_entries():
    _ledger = Ledger(
        {
            "cash": Asset(debits=[], credits=[]),
            "equity": Capital(debits=[], credits=[]),
        }
    )
    _, failed = unsafe_post_entries(_ledger, [Entry("", "", 0)])
    assert failed == [Entry("", "", 0)]


def test_create_ledger_again():
    (
        Chart(
            assets=["cash", "goods", "ppe"],
            equity=["equity", "re"],
            income=["sales"],
            expenses=["cogs", "sga"],
        )
        .offset("ppe", ["depreciation"])
        # https://stripe.com/docs/revenue-recognition/methodology
        .offset("sales", ["refunds", "voids"])
        .ledger()
    ) == {
        "cash": Asset(debits=[], credits=[]),
        "goods": Asset(debits=[], credits=[]),
        "ppe": Asset(debits=[], credits=[]),
        "cogs": Expense(debits=[], credits=[]),
        "sga": Expense(debits=[], credits=[]),
        "equity": Capital(debits=[], credits=[]),
        "sales": Income(debits=[], credits=[]),
        "depreciation": ContraAsset(debits=[], credits=[]),
        "refunds": ContraIncome(debits=[], credits=[]),
        "voids": ContraIncome(debits=[], credits=[]),
        "null": NullAccount(debits=[], credits=[]),
        "re": RetainedEarnings(debits=[], credits=[]),
        "current_profit": IncomeSummaryAccount(debits=[], credits=[]),
    }


def test_make_ledger_with_netting():
    chart = Chart(
        assets=["ppe"], expenses=[], equity=["shares"], liabilities=[], income=["sales"]
    )
    chart.contra_accounts = {
        "sales": ["refunds", "voids"],
        "shares": ["treasury_shares"],
        "ppe": ["depreciation"],
    }
    assert chart.ledger() == {
        "ppe": Asset(
            debits=[],
            credits=[],
        ),
        "shares": Capital(
            debits=[],
            credits=[],
        ),
        "re": RetainedEarnings(debits=[], credits=[]),
        "sales": Income(debits=[], credits=[]),
        "refunds": ContraIncome(debits=[], credits=[]),
        "voids": ContraIncome(debits=[], credits=[]),
        "treasury_shares": ContraCapital(debits=[], credits=[]),
        "depreciation": ContraAsset(debits=[], credits=[]),
        "current_profit": IncomeSummaryAccount(debits=[], credits=[]),
        "null": NullAccount(debits=[], credits=[]),
    }



import pytest  # type: ignore
from pytest import fixture  # type: ignore

from abacus.engine.base import Entry
from abacus.engine.chart import Chart
from abacus.engine.ledger import Ledger


@fixture
def chart():
    return Chart(
        assets=["cash", "ar", "goods"],
        equity=["equity"],
        income=["sales"],
        expenses=["cogs", "sga"],
        contra_accounts={"sales": ["refunds", "voids"]},
    )


@fixture
def ledger(chart):
    return Ledger.new(chart).post_many(
        [
            Entry("cash", "equity", 1000),
            Entry("ar", "sales", 250),
            Entry("refunds", "ar", 50),
        ]
    )


@pytest.fixture
def entries():
    e1 = Entry(debit="cash", credit="equity", amount=1000)
    e2 = Entry(debit="goods", credit="cash", amount=250)
    e3 = Entry(debit="cogs", credit="goods", amount=200)
    e4 = Entry(debit="cash", credit="sales", amount=400)
    e5 = Entry(debit="sga", credit="cash", amount=50)
    return [e1, e2, e3, e4, e5]



"""Closing accounts at period end."""
from dataclasses import dataclass
from typing import List, Type

from abacus.engine.accounts import (
    Asset,
    Capital,
    Expense,
    Income,
    Liability,
    RegularAccount,
)
from abacus.engine.base import AccountName, Entry
from abacus.engine.chart import Chart
from abacus.engine.ledger import Ledger

__all__ = [
    "close_contra_income",
    "close_contra_expense",
    "make_closing_entries",
    "flush_permanent_accounts",
]


def flush_permanent_accounts(chart: Chart, ledger: Ledger) -> List[Entry]:
    def f(cls):
        return _flush(chart, ledger, cls)

    return f(Asset) + f(Capital) + f(Liability)


@dataclass
class ClosingEntries:
    """Closing entries for accounts related to income and expenses."""

    closing_contra_income: List[Entry]
    closing_contra_expense: List[Entry]
    closing_income: List[Entry]
    closing_expense: List[Entry]
    closing_isa: Entry

    @property
    def closing_contra_accounts(self) -> List[Entry]:
        return self.closing_contra_income + self.closing_contra_expense

    @property
    def closing_income_expense_isa(self) -> List[Entry]:
        return self.closing_income + self.closing_expense + [self.closing_isa]

    def all(self):
        return self.closing_contra_accounts + self.closing_income_expense_isa


def make_closing_entries(chart: Chart, ledger: Ledger) -> ClosingEntries:
    """Create a data structure holding closing entries.

    Steps taken are the following:
    - close contra accounts to income and expense accounts
    - close income and expense accounts
    - close income summary account to retained earnings.
    """
    dummy_ledger = ledger.condense()
    a = close_contra_income(chart, dummy_ledger)
    b = close_contra_expense(chart, dummy_ledger)
    dummy_ledger.post_many(a + b)
    c = close_income_to_isa(chart.income_summary_account, dummy_ledger)
    d = close_expenses_to_isa(chart.income_summary_account, dummy_ledger)
    dummy_ledger.post_many(c + d)
    e = close_isa(
        isa=chart.income_summary_account,
        re=chart.retained_earnings_account,
        ledger=dummy_ledger,
    )
    return ClosingEntries(
        closing_contra_income=a,
        closing_contra_expense=b,
        closing_income=c,
        closing_expense=d,
        closing_isa=e,
    )


def _flush(chart: Chart, ledger: Ledger, cls: Type[RegularAccount]) -> List[Entry]:
    """Make a list of entries for transferring balances for accounts of a certain type.
    The type of accounts is contra accounts for *cls*, e.g. when provided Income,
    the entries will transfer balances from ContraIncome accounts to Income accounts.
    These entries are used to close contra accounts."""
    return [
        ledger[contra_account_name].transfer(contra_account_name, regular_account_name)  # type: ignore
        for regular_account_name, contra_account_name in chart.contra_account_pairs(cls)
    ]


def close_contra_income(chart: Chart, ledger: Ledger) -> List[Entry]:
    """Make entries for closing contra income accounts (eg discounts, refunds made to clients).
    Balance transfered from contra account to corresponding income account.
    The income accounts will hold net value after these entries are posted.
    """
    return _flush(chart, ledger, Income)


def close_contra_expense(chart: Chart, ledger: Ledger) -> List[Entry]:
    """Make entries for closing contra expense accounts (eg discounts received).
    Balance transfered from contra account to corresponding expense account.
    The expense account will hold net value after these entries are posted.
    """
    return _flush(chart, ledger, Expense)


def close_income_to_isa(isa: AccountName, ledger: Ledger) -> List[Entry]:
    return [
        Entry(account_name, isa, t_account.balance())
        for account_name, t_account in ledger.subset(Income).items()
    ]


def close_expenses_to_isa(isa: AccountName, ledger: Ledger) -> List[Entry]:
    return [
        Entry(isa, account_name, t_account.balance())
        for account_name, t_account in ledger.subset(Expense).items()
    ]


def close_isa(isa: AccountName, re: AccountName, ledger: Ledger) -> Entry:
    """Make entry to close income summary account ("isa") and move its balance to retained earnings."""
    amount = ledger[isa].balance()
    return Entry(isa, re, amount)



"""Income statement and balance sheet reports, created from ledger."""
from dataclasses import dataclass
from typing import Dict, List, Tuple

from pydantic import BaseModel  # type: ignore
from rich.console import Console  # type: ignore
from rich.table import Table  # type: ignore
from rich.text import Text  # type: ignore

from abacus.engine.accounts import (
    Asset,
    Capital,
    CreditAccount,
    DebitAccount,
    Expense,
    Income,
    Liability,
)
from abacus.engine.base import AccountName, Amount


def total(self: dict) -> Amount:
    return sum(self.values())


# TODO: make this class ABC
class Report:
    def lines(self, attributes: List[str]) -> List["Line"]:
        """Return list of lines for a given subset attributes.

        Allows to create a report with multiple columns, for example,
        capital and liabilities on the left side of balance sheet"""
        lines: List[Line] = []
        for attr in attributes:
            dict_ = getattr(self, attr)
            # the header line shows sum of the items
            lines.append(HeaderLine(text=attr, value=str(total(dict_))))
            for name, value in dict_.items():
                lines.append(AccountLine(name, str(value)))
        return lines

    def view(self, rename_dict: Dict[str, str]) -> str:
        raise NotImplementedError

    def print(self, rename_dict: Dict[str, str]) -> None:
        print(self.view(rename_dict))


class BalanceSheet(BaseModel, Report):
    assets: Dict[AccountName, Amount]
    capital: Dict[AccountName, Amount]
    liabilities: Dict[AccountName, Amount]

    @classmethod
    def new(cls, ledger):
        return BalanceSheet(
            assets=ledger.subset(Asset).balances(),
            capital=ledger.subset(Capital).balances(),
            liabilities=ledger.subset(Liability).balances(),
        )

    def view(self, rename_dict) -> str:
        return view_balance_sheet(self, rename_dict)

    def print_rich(self, rename_dict: Dict[str, str]) -> None:
        left, right = left_and_right(self, rename_dict)
        table = make_balance_sheet_table("Balance sheet", 80)
        for line_1, line_2 in zip(left, right):
            a, b = unpack(line_1)
            c, d = unpack(line_2)
            table.add_row(a, b, c, d)
        Console().print(table)


class IncomeStatement(BaseModel, Report):
    income: Dict[AccountName, Amount]
    expenses: Dict[AccountName, Amount]

    @classmethod
    def new(cls, ledger) -> "IncomeStatement":
        return IncomeStatement(
            income=ledger.subset(Income).balances(),
            expenses=ledger.subset(Expense).balances(),
        )

    def current_profit(self):
        return total(self.income) - total(self.expenses)

    def view(self, rename_dict) -> str:
        return view_income_statement(self, rename_dict)

    def print_rich(self, rename_dict: Dict[str, str]) -> None:
        print_income_statement_rich(self, rename_dict)


def clean(s: str, rename_dict: Dict) -> str:
    return rename_dict.get(s, s).replace("_", " ").strip().capitalize()


@dataclass
class Line:
    text: str
    value: str

    def rename(self, rename_dict: Dict[str, str]):
        self.text = clean(self.text, rename_dict)
        return self


class HeaderLine(Line):
    pass


class AccountLine(Line):
    pass


class EmptyLine(Line):
    pass


def left_and_right(
    report: BalanceSheet, rename_dict: Dict[str, str]
) -> Tuple[List[Line], List[Line]]:
    left = report.lines(["assets"])
    right = report.lines(["capital", "liabilities"])
    # rename lines
    left = [line.rename(rename_dict) for line in left]
    right = [line.rename(rename_dict) for line in right]
    # make left and right same number of lines
    n = max(len(left), len(right))
    left += [EmptyLine("", "")] * (n - len(left))
    right += [EmptyLine("", "")] * (n - len(right))
    # add end line
    h1 = HeaderLine("Total:", str(total(report.assets)))
    left.append(h1)
    h2 = HeaderLine("Total:", str(total(report.capital) + total(report.liabilities)))
    right.append(h2)
    return left, right


def view_balance_sheet(report: BalanceSheet, rename_dict: Dict[str, str]) -> str:
    left, right = left_and_right(report, rename_dict)
    a, b = to_columns(left)
    col1 = a.align_left(".").add_right("... ") + b.align_right()
    c, d = to_columns(right)
    col2 = c.align_left(".").add_right("... ") + d.align_right()
    return (col1.add_space(2) + col2).printable()


def red(b: str) -> Text:
    """Make digit red if negative."""
    v = Amount(b)
    if v and v < 0:
        return Text(b, style="red")
    else:
        return Text(b)


def bold(s: Text | str) -> Text:
    if isinstance(s, Text):
        s.stylize("bold")
        return s
    else:
        return Text(str(s), style="bold")


def unpack(line: Line) -> Tuple[Text, Text]:
    """Convert Line to rich.Text tuple"""
    match line:
        case HeaderLine(a, b):
            return bold(a), bold(red(b))
        case AccountLine(a, b):
            return (Text("  " + a), red(b))
        case EmptyLine(a, b):
            return Text(""), Text("")
        case _:
            raise ValueError(line)


def income_statement_lines(report: IncomeStatement, rename_dict: Dict[str, str]):
    left = report.lines(["income", "expenses"])
    # rename lines
    left = [line.rename(rename_dict) for line in left]
    # add end line
    h1 = HeaderLine("Current profit:", str(report.current_profit()))
    left.append(h1)
    return left


def view_income_statement(report: IncomeStatement, rename_dict: Dict[str, str]) -> str:
    left = income_statement_lines(report, rename_dict)
    a, b = to_columns(left)
    col = a.align_left(fill_char=".").add_right("... ") + b.align_right(fill_char=" ")
    return col.printable()


def print_income_statement_rich(report: IncomeStatement, rename_dict: Dict[str, str]):
    left = income_statement_lines(report, rename_dict)
    table = Table(title="Income Statement", box=None, width=80, show_header=False)
    table.add_column("")
    table.add_column("", justify="right", style="green")
    for line in left:
        a, b = unpack(line)
        table.add_row(a, b)
    Console().print(table)


def offset(line: Line) -> str:
    match line:
        case HeaderLine(a, _):
            return a.upper()
        case AccountLine(a, _):
            return "  " + a
        case EmptyLine(_, _):
            return ""
    raise TypeError  # mypy wanted it


def to_columns(lines: List[Line]) -> Tuple["Column", "Column"]:
    return Column(list(map(offset, lines))), Column([line.value for line in lines])


@dataclass
class Column:
    strings: List[str]

    @property
    def width(self):
        return max(len(s) for s in self.strings)

    def align_left(self, fill_char=" "):
        return Column([s.ljust(self.width, fill_char) for s in self.strings])

    def align_right(self, fill_char=" "):
        return Column([s.rjust(self.width, fill_char) for s in self.strings])

    def align_center(self):
        return Column([s.center(self.width) for s in self.strings])

    def empty(self, n: int = 1):
        return self.refill(" " * n)

    def add_right(self, string: str):
        return self + self.refill(string)

    def add_space(self, n: int = 1):
        return self + self.empty(n)

    def add_space_left(self, n: int = 1):
        return self.empty(n) + self

    def refill(self, text):
        return Column([text] * len(self.strings))

    def merge(self, column):
        return Column([a + b for a, b in zip(self.strings, column.strings)])

    def __add__(self, column: "Column"):
        return self.merge(column)

    def header(self, text):
        return Column([text.center(self.width)] + self.strings)

    def printable(self):
        return "\n".join(self.strings)


def yield_tuples_for_trial_balance(chart, ledger):
    def t(account_name):
        return chart.get_account_type(account_name).__name__

    for account_name, t_account in ledger.items():
        if isinstance(t_account, DebitAccount):
            yield account_name, t(account_name), t_account.balance(), 0
    for account_name, t_account in ledger.items():
        if isinstance(t_account, CreditAccount):
            yield account_name, t(account_name), 0, t_account.balance()


def nth(data, n, f=lambda x: x):
    return Column([f(d[n]) for d in data])


def view_trial_balance(chart, ledger) -> str:
    data = list(yield_tuples_for_trial_balance(chart, ledger))
    col_1 = (
        Column([chart.compose_name_long(d[0]) for d in data])
        .align_left(".")
        .add_space(1)
        .header("Account")
    )
    col_2 = (
        nth(data, 2, str).align_right().add_space_left(2).header("Debit").add_space(2)
    )
    col_3 = nth(data, 3, str).align_right().add_space_left(2).header("Credit")
    return (col_1 + col_2 + col_3).printable()


def make_balance_sheet_table(title, width) -> Table:
    table = Table(title=title, box=None, width=width, show_header=False)
    table.add_column("")
    table.add_column("", justify="right", style="green")
    table.add_column("")
    table.add_column("", justify="right", style="green")
    return table



from pytest import fixture  # type: ignore

from abacus.engine.accounts import (
    Asset,
    Capital,
    ContraAsset,
    ContraIncome,
    Expense,
    Income,
    IncomeSummaryAccount,
    Liability,
    RetainedEarnings,
)
from abacus.engine.base import Amount, Entry
from abacus.engine.chart import Chart
from abacus.engine.closing import make_closing_entries
from abacus.engine.ledger import Ledger


@fixture
def entries():
    doc_ = """
cash,equity,1000
goods,cash,800
cogs,goods,700
ar,sales,899
refunds,ar,89
cashback,ar,10
cash,ar,400
sga,cash,50"""

    def make(line):
        a, b, c = line.split(",")
        return Entry(a.strip(), b.strip(), Amount(c))

    return [make(line) for line in doc_.strip().split("\n")]


@fixture
def chart2():
    return Chart(
        assets=["cash", "goods", "ar"],
        equity=["equity"],
        income=["sales"],
        expenses=["cogs", "sga"],
        contra_accounts={"sales": ["refunds", "cashback"]},
    )


@fixture
def ledger_before_close(chart2, entries):
    return Ledger.new(chart2).post_many(entries)


def test_len_closing_entries(chart2, ledger_before_close):
    closing_entries = make_closing_entries(chart2, ledger_before_close)
    assert len(closing_entries.all()) == 6


def test_re_after_close(chart2, ledger_before_close):
    assert ledger_before_close.close(chart2).balances()["re"] == 50


def test_balances_after(chart2, ledger_before_close):
    assert ledger_before_close.close_some(chart2).subset(
        [Income, Expense]
    ).balances() == {
        "cogs": 700,
        "sga": 50,
        "sales": 800,
    }


def test_closing_entries():
    chart = (
        Chart(
            assets=["cash", "receivables", "goods_for_sale", "ppe"],
            expenses=["cogs", "sga", "depreciation_expense"],
            equity=["equity"],
            liabilities=["dividend_due", "payables"],
            income=["sales"],
        )
        .offset("ppe", ["depreciation"])
        .offset("sales", ["discount", "returns"])
    )

    ledger = Ledger(
        {
            "cash": Asset(debits=[4000, 620], credits=[3000, 250, 40, 25, 50]),
            "goods_for_sale": Asset(debits=[250], credits=[180]),
            "ppe": Asset(
                debits=[3000],
                credits=[],
            ),
            "cogs": Expense(debits=[180], credits=[]),
            "sga": Expense(debits=[50], credits=[]),
            "depreciation_expense": Expense(debits=[250], credits=[]),
            "equity": Capital(debits=[], credits=[4000]),
            "re": RetainedEarnings(debits=[], credits=[]),
            "divp": Liability(debits=[], credits=[]),
            "sales": Income(
                debits=[],
                credits=[620],
            ),
            "depreciation": ContraAsset(debits=[], credits=[250]),
            "discount": ContraIncome(debits=[40, 25], credits=[]),
            "returns": ContraIncome(debits=[], credits=[]),
            "current_profit": IncomeSummaryAccount(debits=[], credits=[]),
        }
    )

    assert list(make_closing_entries(chart, ledger).all()) == [
        Entry(debit="sales", credit="discount", amount=65),
        Entry(debit="sales", credit="returns", amount=0),
        Entry(debit="sales", credit="current_profit", amount=555),
        Entry(debit="current_profit", credit="cogs", amount=180),
        Entry(debit="current_profit", credit="sga", amount=50),
        Entry(debit="current_profit", credit="depreciation_expense", amount=250),
        Entry(debit="current_profit", credit="re", amount=75),
    ]



import pytest  # type: ignore

from abacus.engine.accounts import Asset, DebitAccount, Income
from abacus.engine.base import AbacusError
from abacus.engine.chart import Chart
from abacus.engine.ledger import Ledger


def test_empty_chart_constructor():
    assert Chart().dict() == dict(
        assets=[],
        expenses=[],
        equity=[],
        liabilities=[],
        income=[],
        retained_earnings_account="re",
        income_summary_account="current_profit",
        null_account="null",
        contra_accounts={},
        names={"re": "Retained earnings"},
        operations={},
        codes={},
    )


def test_chart_duplicates():
    with pytest.raises(AbacusError):
        Chart(assets=["cash"]).offset("cash", "cash").validate()


def test_chart_offset_does_not_exist():
    with pytest.raises(AbacusError):
        Chart().offset("account", "contra_account")


def test_chart_offset():
    assert Chart(income=["sales"]).offset("sales", "refunds").contra_accounts == dict(
        sales=["refunds"]
    )


def test_accounts_names_all(chart):
    assert chart.account_names_all() == [
        "cash",
        "ar",
        "goods",
        "cogs",
        "sga",
        "equity",
        "sales",
        "refunds",
        "voids",
        "re",
        "current_profit",
        "null",
    ]


def test_contra_account_pairs(chart):
    assert list(chart.contra_account_pairs(Income)) == [
        ("sales", "refunds"),
        ("sales", "voids"),
    ]


def test_account_names(chart):
    assert chart.account_names(Asset) == ["cash", "ar", "goods"]


def test_chart_offset_many():
    assert Chart(income=["sales"]).offset(
        "sales", ["refunds", "voids"]
    ).contra_accounts == dict(sales=["refunds", "voids"])


def test_chart_get_name():
    assert Chart().get_name("_dividends_due") == "Dividends due"


def test_chart_set_name_get_name():
    assert Chart().set_name("cos", "Cost of sales").get_name("cos") == "Cost of sales"


def test_is_debit_account():
    acc = Ledger.new(
        Chart(
            assets=["cash"],
            equity=[],
            expenses=[],
            liabilities=[],
            income=[],
        )
    )["cash"]
    assert isinstance(acc, DebitAccount)


def test_offset():
    chart = Chart(
        assets=[],
        equity=[],
        expenses=[],
        liabilities=[],
        income=["sales"],
    ).offset("sales", "refund")
    assert chart.contra_accounts == {"sales": ["refund"]}
    chart.offset("sales", "voids")
    assert chart.contra_accounts == {"sales": ["refund", "voids"]}


chart = (
    Chart(
        assets=["cash", "receivables", "goods_for_sale", "ppe"],
        expenses=["cogs", "sga", "depreciation_expense"],
        equity=["equity"],
        liabilities=["dividend_due", "payables"],
        income=["sales"],
    )
    .offset("ppe", ["depreciation"])
    .offset("sales", ["discount", "returned"])
)


def test_invalid_chart_with_duplicate_key():
    with pytest.raises(AbacusError):
        Chart(
            assets=["cash"],
            expenses=[],
            equity=["cash"],
            liabilities=[],
            income=[],
        ).validate()


def test_invalid_chart_with_non_existent_contra_account_name():
    with pytest.raises(AbacusError):
        Chart(
            assets=["cash", "goods"],
            expenses=["cogs"],
            equity=["capital"],
            liabilities=["loan"],
            income=["sales"],
        ).offset("sssalessssss", ["refunds"])


def test_account_names_method(chart):
    assert chart.account_names(Asset) == ["cash", "ar", "goods"]


def test_creation():
    (
        Chart(
            assets=["cash", "ar", "goods", "ppe"],
            equity=["equity", "retained_earnings"],
            income=["sales"],
            liabilities=["ap"],
            expenses=["cogs", "sga"],
        )
        .offset("ppe", ["depreciation"])
        .offset("sales", ["refunds", "voids"])
        .set_name("ppe", "Property, plant, equipment")
        .set_name("goods", "Inventory (goods for sale)")
        .set_name("ar", "Accounts receivable")
        .set_name("ap", "Accounts payable")
        .set_name("cogs", "Cost of goods sold")
        .set_name("sga", "Selling, general and adm. expenses")
    )
    assert 1



import pytest  # type: ignore

from abacus import BalanceSheet, Chart, Entry, IncomeStatement, Ledger


def test_create_balance_sheet():
    chart = Chart(assets=["cash"], equity=["equity"])
    ledger = Ledger.new(chart).post(Entry("cash", "equity", 1000))
    assert ledger.balance_sheet(chart) == BalanceSheet(
        assets={"cash": 1000}, capital={"equity": 1000, "re": 0}, liabilities={}
    )


@pytest.fixture
def chart0():
    return Chart(
        assets=["cash", "receivables", "goods_for_sale"],
        expenses=["cogs", "sga"],
        equity=["equity"],
        liabilities=["divp", "payables"],
        income=["sales"],
    )


@pytest.fixture
def entries0():
    e1 = Entry(debit="cash", credit="equity", amount=1000)
    e2 = Entry(debit="goods_for_sale", credit="cash", amount=250)
    e3 = Entry(credit="goods_for_sale", debit="cogs", amount=200)
    e4 = Entry(credit="sales", debit="cash", amount=400)
    e5 = Entry(credit="cash", debit="sga", amount=50)
    return [e1, e2, e3, e4, e5]


def test_income_statement(chart0, entries0):
    income_st = chart0.ledger().post_many(entries0).income_statement(chart0)
    assert income_st == IncomeStatement(
        income={"sales": 400}, expenses={"cogs": 200, "sga": 50}
    )


def test_balance_sheet_with_close(chart0, entries0):
    balance_st = chart0.ledger().post_many(entries0).balance_sheet(chart0)
    assert balance_st == BalanceSheet(
        assets={"cash": 1100, "receivables": 0, "goods_for_sale": 50},
        capital={"equity": 1000, "re": 150},
        liabilities={"divp": 0, "payables": 0},
    )


def test_balances(chart0, entries0):
    assert chart0.ledger().post_many(entries0).close(chart0).balances() == {
        "cash": 1100,
        "receivables": 0,
        "goods_for_sale": 50,
        "cogs": 0,
        "sga": 0,
        "equity": 1000,
        "re": 150,
        "divp": 0,
        "payables": 0,
        "sales": 0,
        "current_profit": 0,
        "null": 0,
    }



from abacus.engine.accounts import Asset, ContraIncome, CreditAccount, DebitAccount


def test_asset_topup():
    assert Asset([10], []).topup(10) == Asset([10, 10], [])


def test_asset_condense():
    assert Asset([10, 10], []).condense() == Asset([20], [])


def test_split_on_caps():
    assert DebitAccount([], []).split_on_caps() == "Debit Account"


def test_debit_credit():
    da = DebitAccount([], [])
    da.debit(10)
    da.credit(2)
    assert da == DebitAccount(debits=[10], credits=[2])


def test_account_safe_copy():
    da = DebitAccount(debits=[10], credits=[2])
    da2 = da.deep_copy()
    da2.debit(8)
    da2.credit(15)
    assert da2 == DebitAccount(debits=[10, 8], credits=[2, 15])
    assert da == DebitAccount(debits=[10], credits=[2])


def test_balance_on_DebitAccount():
    assert DebitAccount([10, 10], [5]).balance() == 10 + 10 - 5


def test_balance_on_CreditAccount():
    assert CreditAccount([2, 2], [5, 4]).balance() == 5 + 4 - 2 - 2


def test_ContraIncome():
    assert ContraIncome([8], []).balance() == 8
